"""
Takes a list of server/class/devices (optionally with wildcards)
and prints the current configuration for those in JSON dsconfig format.

$ python -m dsconfig.dump server:TangoTest/1 > result.json
$ python -m dsconfig.dump device:sys/tg_test/1 device:sys/tg_test/2
...

"""

from __future__ import absolute_import
from __future__ import print_function
import tango

from .appending_dict import SetterDict
from .tangodb import get_servers_with_filters, get_classes_properties


def get_db_data(db, patterns=None, class_properties=False, **options):
    # dump TANGO database into JSON. Optionally filter which things to include
    # (currently only "positive" filters are possible; you can say which
    # servers/classes/devices to include, but you can't exclude selectively)
    # By default, dserver devices aren't included!

    dbproxy = tango.DeviceProxy(db.dev_name())
    data = SetterDict()

    if not patterns:
        # the user did not specify a pattern, so we will dump *everything*
        servers = get_servers_with_filters(
            dbproxy, **options)
        data.servers.update(servers)
        if class_properties:
            classes = get_classes_properties(dbproxy)
            data.classes.update(classes)
    else:
        # go through all patterns and fill in the data
        for pattern in patterns:
            prefix, pattern = pattern.split(":")
            kwargs = {prefix: pattern}
            kwargs.update(options)
            servers = get_servers_with_filters(dbproxy, **kwargs)
            data.servers.update(servers)
            if class_properties:
                classes = get_classes_properties(dbproxy, server=pattern)
                data.classes.update(classes)
    return data.to_dict()


def main():
    import json
    from optparse import OptionParser

    usage = "Usage: %prog [term:pattern term2:pattern2...]"
    parser = OptionParser(usage=usage)
    parser.add_option("-p", "--no-properties", dest="properties",
                      action="store_false", default=True,
                      help="Exclude device properties")
    parser.add_option("-a", "--no-attribute-properties",
                      dest="attribute_properties",
                      action="store_false", default=True,
                      help="Exclude attribute properties")
    parser.add_option("-l", "--no-aliases", dest="aliases",
                      action="store_false", default=True,
                      help="Exclude device aliases")
    parser.add_option("-d", "--dservers", dest="dservers",
                      action="store_true", help="Include DServer devices")
    parser.add_option("-s", "--subdevices", dest="subdevices",
                      action="store_true", default=False,
                      help="Include __SubDevices property")
    parser.add_option("-c", "--class-properties",
                      dest="class_properties",
                      action="store_true", default=False,
                      help="Include class properties")

    options, args = parser.parse_args()

    db = tango.Database()
    dbdata = get_db_data(db, args,
                         properties=options.properties,
                         class_properties=options.class_properties,
                         attribute_properties=options.attribute_properties,
                         aliases=options.aliases, dservers=options.dservers,
                         subdevices=options.subdevices)
    print((json.dumps(dbdata, ensure_ascii=False, indent=4, sort_keys=True)))


if __name__ == "__main__":
    main()



from __future__ import absolute_import
import re

from .appending_dict import merge


def filter_nested_dict(node, pattern, depth, level=0, invert=False):
    """
    Filter the parts of a nested dict where keys match regex pattern,
    at the given depth.
    """
    if level == depth:
        return dict((key, value) for key, value in list(node.items())
                    if (not invert and pattern.match(key)) or
                    (invert and not pattern.match(key)))
    else:
        dupe_node = {}
        for key, val in list(node.items()):
            cur_node = filter_nested_dict(val, pattern, depth, level + 1,
                                          invert)
            if cur_node:
                dupe_node[key] = cur_node
        return dupe_node or None


def filter_config(data, filters, levels, invert=False):
    """
    Filter the given config data according to a list of filters.
    May be a positive filter (i.e. includes only matching things)
    or inverted (i.e. includes everything that does not match).
    The _levels_ argument is used to find at what depth in the data
    the filtering should happen.
    """

    filtered = data if invert else {}
    for fltr in filters:
        try:
            what, regex = fltr.split(":")
            if what == "server" and "/" in regex:
                # special case the "server/instance" syntax to return
                # only the specific instance in the server
                srv, inst = [re.compile(r, flags=re.IGNORECASE)
                             for r in regex.split("/")]
                servers = filter_nested_dict(data, srv, 0)
                for k, v in list(servers.items()):
                    tmp = filter_nested_dict(v, inst, 0)
                    if tmp:
                        filtered[k] = tmp
                continue
            depth = levels[what]
            pattern = re.compile(regex, flags=re.IGNORECASE)
        except (ValueError, IndexError):
            raise ValueError(
                "Bad filter '%s'; should be '<term>:<regex>'" % fltr)
        except KeyError:
            raise ValueError("Bad filter '%s'; term should be one of: %s"
                             % (fltr, ", ".join(list(levels.keys()))))
        except re.error as e:
            raise ValueError("Bad regular expression '%s': %s" % (fltr, e))
        if invert:
            filtered = filter_nested_dict(filtered, pattern, depth,
                                          invert=True)
        else:
            tmp = filter_nested_dict(data, pattern, depth)
            if tmp:
                merge(filtered, tmp)
    return filtered



"""
Routines for reading an Excel file containing server, class and device definitions,
producing a file in the TangoDB JSON format.
"""
from __future__ import absolute_import
from __future__ import print_function

import json
import os
import re
import sys
from datetime import datetime

from .appending_dict import AppendingDict
from .tangodb import SPECIAL_ATTRIBUTE_PROPERTIES
from .utils import CaselessDict
from .utils import find_device

# from traceback import format_exc

MODE_MAPPING = CaselessDict({"ATTR": "DynamicAttributes",
                             "CMD": "DynamicCommands",
                             "STATE": "DynamicStates",
                             "STATUS": "DynamicStatus"})

TYPE_MAPPING = CaselessDict({"INT": int,
                             "FLOAT": float})


def get_properties(row):
    """
    Find property definitions on a row
    """

    prop_dict = AppendingDict()

    # "Properties" column
    # The cell value is expected to be on the form
    #   "property1=value1;property2=value2" etc
    # Note: In this case we cannot know the type of the value so we will
    # use the string as it is. This should be safe for e.g. numbers,
    # as long as the format of the string is correct
    if "properties" in row:
        properties = row["properties"]
        try:
            for prop in properties.split(";"):
                name, value = prop.split("=")
                # need to decode the string, otherwise any linebreaks
                # will be escaped.
                value = str(value) # .decode("string-escape")
                # Support inline multiline properties using "\n"
                prop_dict[name.strip()] = [v.strip()
                                           for v in value.split("\n")]
        except ValueError:
            raise ValueError("could not parse Properties")

    # "Property:xyz" and "Property(type):xyz columns
    # The main issue here is that spreadsheet programs treat numeric cells
    # as floats. If the number must be inserterd as an int, use the "(INT)"
    # modifier. There does not seem to be a way to force a numeric cell to
    # be interpreted as a string.
    for col_name, value in list(row.items()):
        match = re.match("property(?:\((.*)\))?:(.*)", col_name, re.IGNORECASE)
        if match and (value is not None):  # protect against zero, false...
            type_, name = match.groups()
            if type_:
                convert = TYPE_MAPPING[type_]
                values = [convert(value)]
            else:
                value = str(value)
                values = [v.strip() for v in value.split("\n")]
            prop_dict[name] = values

    return prop_dict


def get_attribute_properties(row):
    if "attribute" in row:
        attribute = row["attribute"]
        prop_dict = AppendingDict()
        if "attributeproperties" in row:
            properties = row["attributeproperties"]
            try:
                for prop in properties.split(";"):
                    name, value = prop.split("=")
                    name = name.strip()
                    if name not in SPECIAL_ATTRIBUTE_PROPERTIES:
                        raise ValueError(
                            "'%s' is not a valid attribute property" % name)
                    prop_dict[name.strip()] = [v.strip()
                                               for v in value.split("\n")]
            except ValueError:
                raise ValueError("could not parse AttributeProperties")

        for col_name, value in list(row.items()):
            match = re.match("attrprop:(.*)", col_name, re.IGNORECASE)
            if match and value:
                name, = match.groups()
                name = make_db_name(name.strip())
                if name not in SPECIAL_ATTRIBUTE_PROPERTIES:
                    raise ValueError("'%s' it not a valid attribute property"
                                     % name)
                value = str(value)
                values = [v.strip() for v in value.split("\n")]
                prop_dict[name] = values

        return {attribute: prop_dict}


def get_dynamic(row):
    """
    Find dynamic definitions on a row
    """

    prop_dict = AppendingDict()
    try:
        formula = row["formula"].strip()
        if "type" in row:
            # TODO: Sanity check type?
            formula = "%s(%s)" % (row["type"], formula)
        check_formula(formula)
        mode = str(row["mode"])
        if mode.lower() == "status":
            dyn = formula
        else:
            dyn = "%s=%s" % (row["name"], formula)
        prop_dict[MODE_MAPPING[mode]] = dyn
    except KeyError as e:
        raise ValueError("Problem with formula: %s" % e)

    return prop_dict


def make_db_name(name):
    """
    Convert a Space Separated Name into a lowercase, underscore_separated_name
    """
    return name.strip().lower().replace(" ", "_")


def check_formula(formula):
    """
    Syntax check a dynamic formula.
    """
    compile(formula, "<stdin>", "single")


def check_device_format(devname):
    """
    Verify that a device name is of the correct form (three parts
    separated by slashes, only letters, numbers, dashes and
    underscores allowed.)  Note: We could put more logic here to make
    device names conform to a standard.
    """
    device_pattern = "^[\w-]+/[\w-]+/[\w-]+$"
    if not re.match(device_pattern, devname):
        raise ValueError("device name '%s' not valid" % devname)


def format_server_instance(row):
    """
    Format a server/instance string
    """
    # TODO: handle numeric instance names? They tend to turn up as floats...
    return "%s/%s" % (row["server"], row["instance"])


def convert(rows, definitions, skip=True, dynamic=False, config=False):
    """
    Update a dict of definitions from data
    """

    errors = []
    column_names = rows[0]

    def handle_error(i, msg):
        if skip:
            errors.append((i, msg))
        else:
            raise

    for i, row_ in enumerate(rows[1:]):
        try:
            # The plan is to try to find all information on the
            # line, raising exceptions if there are unrecoverable
            # problems. Those are caught and reported.

            # Filter out empty columns
            row = CaselessDict(dict((str(name), col)
                                    for name, col in zip(column_names, row_)
                                    if col not in ("", None)))

            # Skip empty lines
            if not row:
                continue

            # Target of the properties; device or class?
            if "device" in row:
                check_device_format(row["device"])
                if "server" in row:
                    # full device definition
                    # target is "lazily" evaluated, so that we don't create
                    # an empty dict if it turns out there are no members
                    target = lambda: definitions.servers[row["server"]][row["instance"]] \
                        [row["class"]][row["device"]]
                else:
                    # don't know if the device is already defined
                    target = lambda: find_device(definitions, row["device"])[0]
            elif "class" in row:
                # class definition
                target = lambda: definitions.classes[row["class"]]
            else:
                continue

            if dynamic:
                props = get_dynamic(row)
                if props:
                    target().properties = props

            elif config:
                attr_props = get_attribute_properties(row)
                if attr_props:
                    target().attribute_properties = attr_props
            else:
                props = get_properties(row)
                if props:
                    target().properties = props

        except KeyError as ke:
            # handle_error(i, "insufficient data (%s)" % ke)
            pass
        except ValueError as ve:
            handle_error(i, "Error: %s" % ve)
        except SyntaxError as se:
            # TODO: do something here to show more info about the error
            # ex_type, ex, tb = sys.exc_info()
            # "\n".join(format_exc(ex).splitlines()[-3:]
            handle_error(i, "SyntaxError: %s" % se)

    return errors


def print_errors(errors):
    if errors:
        print("%d lines skipped" % len(errors), file=sys.stderr)
        for err in errors:
            line, msg = err
            print("%d: %s" % (line + 1, msg), file=sys.stderr)


def xls_to_dict(xls_filename, pages=None, skip=False):
    """
    Make JSON out of an XLS sheet of device definitions.
    """

    import xlrd

    xls = xlrd.open_workbook(xls_filename)
    definitions = AppendingDict()

    if not pages:  # if no pages given, assume all pages are wanted
        pages = xls.sheet_names()
    else:
        # Always include Dynamics and ParamConfig as they only add stuff
        # to devices already configured anyway.
        if "Dynamics" not in pages and "Dynamics" in xls.sheet_names():
            pages.append("Dynamics")
        if "ParamConfig" not in pages and "ParamConfig" in xls.sheet_names():
            pages.append("ParamConfig")

    for page in pages:

        print("\nPage: %s" % page, file=sys.stderr)
        sheet = xls.sheet_by_name(page)
        rows = [sheet.row_values(i)
                for i in range(sheet.nrows)]
        if not rows:
            continue  # ignore empty pages
        errors = convert(rows, definitions, skip=skip,
                         dynamic=(page == "Dynamics"),
                         config=(page == "ParamConfig"))
        print_errors(errors)

    return definitions


def get_stats(defs):
    """
    Calculate some numbers
    """

    servers = set()
    instances = set()
    classes = set()
    devices = set()

    for server, instances in list(defs.servers.items()):
        servers.add(server)
        instances.update(instances)
        for clsname, devs in list(instances.items()):
            classes.add(clsname)
            for devname, dev in list(devs.items()):
                devices.add(devname)

    return {"servers": len(servers), "instances": len(instances),
            "classes": len(classes), "devices": len(devices)}


def main():
    from optparse import OptionParser

    usage = "usage: %prog [options] XLS [PAGE1, PAGE2, ...]"
    parser = OptionParser(usage=usage)
    parser.add_option("-t", "--test", action="store_true",
                      dest="test", default=False,
                      help="just test, produce no JSON")
    parser.add_option("-q", "--quiet", action="store_false",
                      dest="verbose", default=True,
                      help="don't print errors to stdout")
    parser.add_option("-f", "--fatal", action="store_false",
                      dest="skip", default=True,
                      help="don't skip, treat any parsing error as fatal")

    options, args = parser.parse_args()
    if len(args) < 1:
        sys.exit("You need to give an XLS file as argument.")
    filename = args[0]
    pages = args[1:]

    data = xls_to_dict(filename, pages, skip=options.skip)
    metadata = dict(
        _title="MAX-IV Tango JSON intermediate format",
        _source=os.path.split(sys.argv[1])[-1],
        _version=1,
        _date=str(datetime.now()))
    data.update(metadata)

    if not options.test:
        print(json.dumps(data, indent=4))
        outfile = open('config.json', 'w')
        json.dump(data, outfile, indent=4)

    stats = get_stats(data)

    print(("\n"
           "Total: %(servers)d servers, %(instances)d instances, "
           "%(classes)d classes and %(devices)d devices defined.") % stats, file=sys.stderr)


if __name__ == "__main__":
    main()



"This module concerns the dsconfig JSON file format"
from __future__ import absolute_import
from __future__ import print_function

import json
import sys
import six
from copy import deepcopy, copy
from os import path

import tango

from .appending_dict import SetterDict

SERVERS_LEVELS = {"server": 0, "instance": 1, "class": 2, "device": 3, "property": 5}
CLASSES_LEVELS = {"class": 1, "property": 2}

module_path = path.dirname(path.realpath(__file__))
SCHEMA_FILENAME = path.join(module_path, "schema/schema2.json")  # rename


# functions to decode unicode JSON (pytango does not like unicode strings)

def decode_list(data):
    rv = []
    for item in data:
        if isinstance(item, six.text_type):
            item = str(item.encode('utf-8'))
        elif isinstance(item, list):
            item = decode_list(item)
        elif isinstance(item, dict):
            item = decode_dict(item)
        rv.append(item)
    return rv


def decode_dict(data):
    rv = {}
    for key, value in data.items():
        if isinstance(key, six.text_type):
            key = str(key.encode('utf-8'))
        if isinstance(value, six.text_type):
            value = str(value.encode('utf-8'))
        elif isinstance(value, list):
            value = decode_list(value)
        elif isinstance(value, dict):
            value = decode_dict(value)
        rv[key] = value
    return rv


def validate_json(data):
    """
    Validate that a given dict is of the right form
    """
    try:
        from jsonschema import validate, exceptions
        with open(SCHEMA_FILENAME) as schema_json:
            schema = json.load(schema_json)
        validate(data, schema)
    except ImportError:
        print(("WARNING: 'jsonschema' not installed, could not "
               "validate json file. You're on your own."), file=sys.stderr)
    except exceptions.ValidationError as e:
        print("ERROR: JSON data does not match schema: %s" % e, file=sys.stderr)
        sys.exit(1)


def load_json(f):
    return json.load(f)


def expand_config(config):
    """
    Takes a configuration dict and expands it into the canonical
    format. This currently means that the server instance level is
    split into a server and an instance level.
    """

    expanded = deepcopy(config)
    if "servers" in config:
        for servername in config["servers"]:
            if "/" in servername:
                server, instance = servername.split("/")
                data = expanded["servers"].pop(servername)
                if server in expanded["servers"]:
                    expanded["servers"][server].update({instance: data})
                else:
                    expanded["servers"][server] = {instance: data}
    return expanded


def clean_metadata(data):
    """
    Removes any keys in the data that begin with '_'
    """
    tmp = copy(data)
    for key in list(tmp.keys()):
        if key.startswith("_"):
            tmp.pop(key, None)
    return tmp


def normalize_config(config):
    """
    Take a 'loose' config and return a new config that conforms to the
    DSConfig format.

    Current transforms:

    - server instances (e.g. 'TangoTest/1') are split into a server
      level and an instance level (e.g. 'TangoTest' -> '1'). This is to
      convert "v1" format files to the "v2" format.

    - "devices" toplevel; allows to change *existing* devices by just
      adding them directly to a "devices" key in the config, instead
      of having to list out the server, instance and class (since this
      information can be gotten from the DB.)

    """
    old_config = expand_config(config)
    new_config = SetterDict()
    if "servers" in old_config:
        new_config.servers = old_config["servers"]
    if "classes" in old_config:
        new_config.classes = old_config["classes"]
    if "devices" in old_config:
        db = tango.Database()
        for device, props in list(old_config["devices"].items()):
            try:
                info = db.get_device_info(device)
            except tango.DevFailed as e:
                sys.exit("Can't reconfigure device %s: %s" % (device, str(e[0].desc)))
            srv, inst = info.ds_full_name.split("/")
            new_config.servers[srv][inst][info.class_name][device] = props

    return new_config.to_dict()



"Various functionality for dealing with the TANGO database"

from __future__ import absolute_import
from collections import defaultdict
from itertools import islice

import tango
from dsconfig.utils import green, red, yellow

from .appending_dict import AppendingDict, SetterDict, CaselessDictionary

# These are special properties that we'll ignore for now
PROTECTED_PROPERTIES = [
    "polled_attr", "logging_level", "logging_target"
]

SPECIAL_ATTRIBUTE_PROPERTIES = [
    "label", "format", "unit", "standard_unit", "display_unit",
    "min_value", "min_alarm", "min_warning",
    "max_value", "max_alarm", "max_warning",
    "delta_t", "delta_val",
    "abs_change", "rel_change",
    "event_period", "archive_abs_change", "archive_rel_change",
    "archive_period", "description", "mode",
    "__value", "__value_ts"  # memorized attribute values go here
]


def get_devices_from_dict(dbdict, name=None):
    return [(server_name, instance_name, class_name, device_name)
            for server_name, server in dbdict.items()
            for instance_name, instance in server.items()
            for class_name, clss in instance.items()
            for device_name in clss
            if name is None or device_name.lower() == name.lower()]


def get_servers_from_dict(dbdict):
    servers = set()
    for server, children in dbdict.get("servers", {}).items():
        if "/" in server:
            servers.add(server)
        else:
            for inst in children:
                servers.add(("%s/%s" % (server, inst)).lower())
    return servers


def is_protected(prop, attr=False):
    """
    There are certain properties that need special treatment as they
    are handled in particular ways by Tango. In general, we don't want to
    remove these if they exist, but we do allow overwriting them.
    """
    if attr:
        # Attribute config properties
        return prop.startswith("_") or prop in SPECIAL_ATTRIBUTE_PROPERTIES
    else:
        return prop.startswith("_") or prop in PROTECTED_PROPERTIES


def present_calls(indata, dbdata, dbcalls):
    # WIP!

    def add_device(devname, devinfo):
        msg = ("ADD DEVICE\n" +
               "{info.server_id} / {info.dev_class} / {info.name}"
               .format(info=devinfo))
        return (green, msg)

    def delete_device(devname):
        return red, "DELETE DEVICE\n {0}".format(devname)


def summarise_calls(dbcalls, dbdata):
    """
    A brief summary of the operations performed by a list of DB calls
    """

    methods = [
        "add_device",
        "delete_device",
        "put_device_property",
        "delete_device_property",
        "put_device_attribute_property",
        "delete_device_attribute_property",
        "put_class_property",
        "delete_class_property",
        "put_class_attribute_property",
        "delete_class_attribute_property",
        "put_device_alias",
        "delete_device_alias"
    ]

    old_servers = get_servers_from_dict(dbdata)
    new_servers = set()
    servers = defaultdict(set)
    devices = defaultdict(set)
    counts = defaultdict(int)

    for method, args, kwargs in dbcalls:
        if method == "add_device":
            info = args[0]
            servers[method].add(info.server)
            if not info.server.lower() in old_servers:
                new_servers.add(info.server)
            n = 1
        elif "device_property" in method:
            n = len(args[1])
            devices[method].add(args[0].upper())
        elif "attribute_property" in method:
            n = sum(len(ps) for attr, ps in args[1].items())
            devices[method].add(args[0].upper())
        elif "property" in method:
            n = len(args[1])
        else:
            n = 1
        counts[method] += n

    messages = {
        "add_device": (green, "Add %%d devices to %d servers." %
                       len(servers["add_device"])),
        "delete_device": (red, "Delete %d devices."),
        "put_device_property": (
            yellow, "Add/change %%d device properties in %d devices." %
            len(devices["put_device_property"])),
        "delete_device_property": (
            red, "Delete %%d device properties from %d devices." %
            len(devices["delete_device_property"])),
        "put_device_attribute_property": (
            yellow, "Add/change %%d device attribute properties in %d devices." %
            len(devices["put_device_attribute_property"])),
        "delete_device_attribute_property": (
            red, "Delete %%d device attribute properties from %d devices." %
            len(devices["delete_device_attribute_property"])),
        "put_class_property": (yellow, "Add/change %d class properties."),
        "delete_class_property": (red, "Delete %d class properties."),
        "put_class_attribute_property": (
            yellow, "Add/change %d class attribute properties"),
        "delete_class_attribute_property": (
            red, "Delete %d class attribute properties."),
        "put_device_alias": (green, "Add/change %d device aliases"),
        "delete_device_alias": (red, "Delete %d device aliases")
    }

    summary = []
    if new_servers:
        summary.append(green("Add %d servers." % len(new_servers)))
    for method in methods:
        if method in counts:
            color, message = messages[method]
            summary.append(color(message % counts[method]))

    return summary


def get_device(db, devname, data, skip_protected=True):
    """
    Returns all relevant DB information about a given device:
    alias (if any), properties, attribute properties
    """

    dev = {}

    try:
        alias = db.get_alias_from_device(devname)
        dev["alias"] = alias
    except tango.DevFailed:
        pass

    # Properties
    properties = {}
    db_props = db.get_device_property_list(devname, "*")
    if db_props:
        props = db.get_device_property(devname, list(db_props))
        for prop, value in props.items():
            # We'll ignore "protected" properties unless they are present
            # in the input data (in that case we want to show that
            # they are changed)
            if (not (skip_protected and is_protected(prop))
                    or prop in data.get("properties", {})):
                value = [str(v) for v in value]  # is this safe?
                properties[prop] = value
        dev["properties"] = properties

    # Attribute properties
    # Seems impossible* to get the full list of defined attribute
    # properties through the API so we'll have to make do with
    # the attributes we know about.
    # OTOH, usually if the attribute config changes, it's because
    # a user has set e.g. the format, and then we should not just
    # remove it, right?
    # (* It is possible, just not through the DB API. See the
    #    "DbMySqlSelect" command on the db device.)
    attr_props = CaselessDictionary(data.get("attribute_properties", {}))
    if attr_props:
        attribute_properties = {}
        dbprops = db.get_device_attribute_property(devname,
                                                   list(attr_props.keys()))
        for attr, props in list(dbprops.items()):
            attr_props = dict(
                (prop, [str(v) for v in values])
                for prop, values in props.items()
                # Again, ignore attr_props that are not present in the
                # input data, as we most likely don't want to remove those
                if (not (skip_protected and is_protected(prop, True))
                    or prop in CaselessDictionary(attr_props.get(attr, {})))
            )  # whew!
            if attr_props:
                attribute_properties[attr] = attr_props
        if attribute_properties:
            dev["attribute_properties"] = attribute_properties

    return dev


def get_dict_from_db(db, data, narrow=False, skip_protected=True):
    """
    Takes a data dict, checks if any if the definitions are already
    in the DB and returns a dict describing them.

    By default it includes all devices for each server+class, use the
    'narrow' flag to limit to the devices present in the input data.
    """

    # This is where we'll collect all the relevant data
    dbdict = SetterDict()
    moved_devices = defaultdict(list)

    # Devices that are already defined somewhere else
    for server, inst, clss, device in get_devices_from_dict(
            data.get("servers", {})):
        try:
            devinfo = db.get_device_info(device)
            srvname = "%s/%s" % (server, inst)
            if devinfo.ds_full_name.lower() != srvname.lower():
                moved_devices[devinfo.ds_full_name].append((clss, device))

        except tango.DevFailed:
            pass

    # Servers
    for srvr, insts in list(data.get("servers", {}).items()):
        for inst, classes in insts.items():
            for clss, devs in classes.items():
                devs = CaselessDictionary(devs)
                if narrow:
                    devices = list(devs.keys())
                else:
                    srv_full_name = "%s/%s" % (srvr, inst)
                    devices = db.get_device_name(srv_full_name, clss)
                for device in devices:
                    new_props = devs.get(device, {})
                    db_props = get_device(db, device, new_props, skip_protected)
                    dbdict.servers[srvr][inst][clss][device] = db_props

    # Classes
    for class_name, cls in data.get("classes", {}).items():
        props = list(cls.get("properties", {}).keys())
        for prop, value in db.get_class_property(class_name, props).items():
            if value:
                value = [str(v) for v in value]
                dbdict.classes[class_name].properties[prop] = value

        attr_props = cls.get("attribute_properties")
        if attr_props:
            dbprops = db.get_class_attribute_property(class_name,
                                                      list(attr_props.keys()))
            for attr, props in dbprops.items():
                props = dict((prop, [str(v) for v in values])
                             for prop, values in props.items())
                dbdict.classes[class_name].attribute_properties[attr] = props

    return dbdict.to_dict(), moved_devices


def find_empty_servers(db, data):
    """
    Find any servers in the data that contain no devices, and remove them
    """
    servers = ["%s/%s" % (srv, inst)
               for srv, insts in data["servers"].items()
               for inst in list(insts.keys())]
    return [server for server in servers
            if all(d.lower().startswith('dserver/')
                   for d in db.get_device_class_list(server))]


def get_device_property_values(dbproxy, device, name="*",
                               include_subdevices=False):
    query = ("SELECT name, value "
             "FROM property_device "
             "WHERE device = '%s' AND name LIKE '%s' "
             "ORDER BY count ASC")
    _, result = dbproxy.command_inout("DbMySqlSelect",
                                      query % (device, name.replace("*", "%")))
    data = defaultdict(list)
    for prop, row in zip(result[::2], result[1::2]):
        if prop != "__SubDevices" or include_subdevices:
            data[prop].append(row)
    return data


def get_device_attribute_property_values(dbproxy, device, name="*"):
    query = ("SELECT attribute, name, value "
             "FROM property_attribute_device "
             "WHERE device = '%s' AND name LIKE '%s' "
             "ORDER BY count ASC")
    _, result = dbproxy.command_inout("DbMySqlSelect",
                                      query % (device, name.replace("*", "%")))
    data = AppendingDict()
    for attr, prop, row in zip(result[::3], result[1::3], result[2::3]):
        data[attr][prop] = row
    return data


def get_devices_for_class(dbproxy, clss):
    query = ("SELECT name FROM device WHERE class LIKE '%s'")
    _, result = dbproxy.command_inout("DbMySqlSelect", query % clss.replace("*", "%"))
    return result


def get_devices_by_name_and_class(dbproxy, name, clss="*"):
    query = ("SELECT name FROM device WHERE name LIKE '%s' "
             "AND class LIKE '%s'")
    _, result = dbproxy.command_inout("DbMySqlSelect",
                                      query % (name.replace("*", "%"), clss.replace("*", "%")))
    return result


def nwise(it, n):
    # [s_0, s_1, ...] => [(s_0, ..., s_(n-1)), (s_n, ... s_(2n-1)), ...]
    return list(zip(*[islice(it, i, None, n) for i in range(n)]))


def maybe_upper(s, upper=False):
    if upper:
        return s.upper()
    return s


def get_servers_with_filters(dbproxy, server="*", clss="*", device="*",
                             properties=True, attribute_properties=True,
                             aliases=True, dservers=False,
                             subdevices=False, uppercase_devices=False,
                             timeout=10):
    """
    A performant way to get servers and devices in bulk from the DB
    by direct SQL statements and joins, instead of e.g. using one
    query to get the properties of each device.

    TODO: are there any length restrictions on the query results? In
    that case, use limit and offset to get page by page.
    """

    server = server.replace("*", "%")  # mysql wildcards
    clss = clss.replace("*", "%")
    device = device.replace("*", "%")

    # Collect all info about each device in this dict
    devices = CaselessDictionary()

    # Queries can sometimes take more than de default 3 s, so it's
    # good to increase the timeout a bit.
    # TODO: maybe instead use automatic retry and increase timeout
    # each time?
    dbproxy.set_timeout_millis(timeout * 1000)

    if properties:
        # Get all relevant device properties
        query = (
            "SELECT device, property_device.name, property_device.value"
            " FROM property_device"
            " INNER JOIN device ON property_device.device = device.name"
            " WHERE server LIKE '%s' AND class LIKE '%s' AND device LIKE '%s'")
        if not dservers:
            query += " AND class != 'DServer'"
        if not subdevices:
            query += " AND property_device.name != '__SubDevices'"
        query += " ORDER BY property_device.count ASC"
        _, result = dbproxy.command_inout("DbMySqlSelect",
                                          query % (server, clss, device))
        for d, p, v in nwise(result, 3):
            dev = devices.setdefault(d, {})
            # Make sure to use caseless dicts, since in principle property names
            # should be caseles too.
            props = dev.setdefault("properties", CaselessDictionary())
            value = props.setdefault(p, [])
            value.append(v)

    if attribute_properties:
        # Get all relevant attribute properties
        query = (
            "SELECT device, attribute, property_attribute_device.name,"
            " property_attribute_device.value"
            " FROM property_attribute_device"
            " INNER JOIN device ON property_attribute_device.device ="
            " device.name"
            " WHERE server LIKE '%s' AND class LIKE '%s' AND device LIKE '%s'")
        if not dservers:
            query += " AND class != 'DServer'"
        query += " ORDER BY property_attribute_device.count ASC"
        _, result = dbproxy.command_inout("DbMySqlSelect", query % (server, clss, device))
        for d, a, p, v in nwise(result, 4):
            dev = devices.setdefault(d, {})
            props = dev.setdefault("attribute_properties", CaselessDictionary())
            attr = props.setdefault(a, CaselessDictionary())
            value = attr.setdefault(p, [])
            value.append(v)

    # dump relevant servers
    query = (
        "SELECT server, class, name, alias FROM device"
        " WHERE server LIKE '%s' AND class LIKE '%s' AND name LIKE '%s'")

    if not dservers:
        query += " AND class != 'DServer'"
    _, result = dbproxy.command_inout("DbMySqlSelect", query % (server, clss, device))

    # combine all the information we have
    servers = SetterDict()
    for s, c, d, a in nwise(result, 4):
        try:
            srv, inst = s.split("/")
        except ValueError:
            # Malformed server name? It can happen!
            continue
        device = devices.get(d, {})
        if a and aliases:
            device["alias"] = a
        devname = maybe_upper(d, uppercase_devices)
        servers[srv][inst][c][devname] = device

    return servers


def get_classes_properties(dbproxy, server='*', cls_properties=True,
                           cls_attribute_properties=True, timeout=10):
    """
    Get all classes properties from server wildcard
    """
    # Mysql wildcards
    server = server.replace("*", "%")
    # Change device proxy timeout
    dbproxy.set_timeout_millis(timeout * 1000)
    # Classes output dict
    classes = AppendingDict()
    # Get class properties
    if cls_properties:
        columns = "class", "name", "value", "count"
        columns_query = ", ".join("property_class.%s" % column for column in columns)
        querry = (
            "select DISTINCT %s "
            "FROM property_class "
            "INNER JOIN device "
            "ON property_class.class = device.class "
            "WHERE server like '%s' "
            "AND device.class != 'DServer' "
            "AND device.class != 'TangoAccessControl' "
            "ORDER BY property_class.count ASC")
        _, result = dbproxy.command_inout("DbMySqlSelect", querry % (columns_query, server))
        # Build the output based on: class, property: value
        for c, p, v, _ in nwise(result, len(columns)):
            classes[c].properties[p] = v
    # Get class attribute properties
    if cls_attribute_properties:
        columns = "class", "attribute", "name", "value", "count"
        columns_query = ", ".join("property_attribute_class.%s" % column for column in columns)
        querry = (
            "select DISTINCT %s "
            "FROM property_attribute_class "
            "INNER JOIN device "
            "ON property_attribute_class.class = device.class "
            "WHERE server like '%s' "
            "AND device.class != 'DServer' "
            "AND device.class != 'TangoAccessControl' "
            "ORDER BY property_attribute_class.count ASC")
        _, result = dbproxy.command_inout("DbMySqlSelect", querry % (columns_query, server))
        # Build output: class, attribute, property: value
        for c, a, p, v, _ in nwise(result, len(columns)):
            # some properties are encoded in latin-1; we want utf-8
            if hasattr(v, "decode"):
                decoded_value = v.decode('iso-8859-1').encode('utf8')  # python 2
            else:
                decoded_value = v  # python 3
            classes[c].attribute_properties[a][p] = decoded_value
    # Return classes collection
    return classes



from __future__ import absolute_import
from __future__ import print_function
import json
import sys

import tango

from .utils import (ObjectWrapper, decode_dict, get_dict_from_db,
                    RED, GREEN, YELLOW)


def delete_devices(db, dbdict, server, cls, devices):
    for devname in devices:
        if dbdict.servers[server][cls][devname]:
            db.delete_device(devname)
        else:
            print("no device", devname)


def delete_server(db, dbdict, servername, serverdata):
    for clsname, devices in list(serverdata.items()):
        delete_devices(db, dbdict, servername, clsname, list(devices.keys()))
    try:
        db.delete_server_info(servername)  # what does this do?
        db.delete_server(servername)
    except tango.DevFailed:
        # I'm not sure about this; sometimes deleting servers works
        # and sometimes not; should they be running? What's going on?
        # Anyway, the worst case scenario is that the servers are
        # still there but contain no devices...
        print("Removing server '%s' may have failed." % servername)


def delete_class(db, dbdict, classname):
    props = list(dbdict.classes[classname].properties.keys())
    if props:
        db.delete_class_property(classname, props)
    attr_props = list(dbdict.classes[classname].attribute_properties.keys())
    if attr_props:
        db.delete_class_attribute_property(classname, attr_props)


def main(json_file, write=False, db_calls=False):
    """
    Remove the devices and servers defined in the given file from the DB.
    """

    with open(json_file) as f:
        data = json.load(f, object_hook=decode_dict)

    db = tango.Database()
    dbdict = get_dict_from_db(db, data)  # check the current DB state

    # wrap the database (and fake it if we're not going to write)
    db = ObjectWrapper(db if write else None)

    # delete servers and devices
    for servername, serverdata in list(data.get("servers", {}).items()):
        if servername in dbdict.servers:
            delete_server(db, dbdict, servername, serverdata)

    # delete classes
    for classname, classdata in list(data.get("classes", {}).items()):
        if classname in dbdict.classes:
            delete_class(db, dbdict, classname)

    if db_calls:
        print("\nTango database calls:")
        for method, args, kwargs in db.calls:
            print(method, args)

    if db.calls:
        if write:
            print((
                    RED + "\n*** Data was written to the Tango DB ***"), file=sys.stderr)
        else:
            print(YELLOW + \
                  "\n*** Nothing was written to the Tango DB (use -w) ***", file=sys.stderr)
    else:
        print(GREEN + "\n*** No changes needed in Tango DB ***", file=sys.stderr)


if __name__ == "__main__":
    from optparse import OptionParser

    usage = "Usage: %prog [options] JSONFILE"
    parser = OptionParser(usage=usage)

    parser.add_option("-w", "--write", dest="write", action="store_true",
                      help="write to the Tango DB", metavar="WRITE")
    parser.add_option("-d", "--dbcalls", dest="dbcalls", action="store_true",
                      help="print out all db calls.")

    options, args = parser.parse_args()

    main(args[0], write=options.write, db_calls=options.dbcalls)



from __future__ import absolute_import
from __future__ import print_function
import json
import sys

from jsonschema import validate, exceptions

from .utils import decode_dict

if __name__ == "__main__":
    data_filename, schema_filename = sys.argv[1], sys.argv[2]

    print("Validating '%s' against schema '%s'..." % (data_filename, schema_filename), end=' ')

    with open(data_filename) as data_json:
        data = json.load(data_json, object_hook=decode_dict)

    with open(schema_filename) as schema_json:
        schema = json.load(schema_json)

    try:
        validate(data, schema)
    except exceptions.ValidationError as e:
        print("data does not match schema:")
        print(e)
        sys.exit(1)
    else:
        print("success!")



from __future__ import absolute_import
from __future__ import print_function
from difflib import ndiff

from tango.utils import CaselessDict
from dsconfig.tangodb import get_devices_from_dict
from dsconfig.utils import green, red, yellow

from .appending_dict import SetterDict


def get_device_data(device, mapping, data):
    if device in mapping:
        server, instance, clss = mapping[device]
        return CaselessDict(data["servers"][server][instance][clss]).get(device, {})
    return {}


def property_diff(old, new, indentation=""):
    diff = ndiff(old, new)
    lines = []
    for line in diff:
        if line.startswith("+"):
            lines.append(green("{}{}".format(indentation, line)))
        elif line.startswith("-"):
            lines.append(red("{}{}".format(indentation, line)))
        else:
            lines.append("{}{}".format(indentation, line))
    return "\n".join(lines)


def format_property(value, indentation="", max_lines=10):
    if len(value) > max_lines:
        ending = "\n{}... [{} lines]".format(indentation, len(value) - max_lines)
    else:
        ending = ""
    return "\n".join("{}{}".format(indentation, line)
                     for line in value[:max_lines]) + ending


def get_changes(data, calls):
    """
    Combine a list of database calls into "changes" that can
    be more easily turned into a readable representation
    """

    devices = get_devices_from_dict(data["servers"])
    device_mapping = CaselessDict(
        (device, (server, instance, clss))
        for server, instance, clss, device in devices
    )
    classes = data.get("classes", {})

    # The idea is to first go through all database calls and collect
    # the changes per device. Then we can print it all out in a more
    # readable format since it will all be collected per device.
    #
    # Example of resulting dict:
    # {
    #     "sys/tg_test/1": {
    #         "server": "TangoTest",
    #         "instance": "test",
    #         "device_class": "TangoTest",
    #         "properties": {
    #             # added property has no old_value
    #             "hej": {
    #                 "value": ["a", "b", "c"]
    #             },
    #             # removed property has no value
    #             "svej": {
    #                 "old_value": ["dsaodkasokd"]
    #             },
    #             # changed property has both
    #             "flepp": {
    #                 "old_value": ["1", "3"],
    #                 "value": ["1", "2", "3"]
    #             }
    #         }
    #     }
    # }

    changes = {
        "devices": SetterDict(),
        "classes": SetterDict()
    }

    # Go through all database calls and store the changes
    for method, args, kwargs in calls:

        if method == "put_device_alias":
            device, alias = args
            if device in device_mapping:
                old_alias = get_device_data(
                    device, device_mapping, data).get("alias")
            else:
                old_alias = None
            changes["devices"][device].update({
                "alias": {"old_value": old_alias,
                          "value": alias}
            })

        elif method == "add_device":
            info, = args
            changes["devices"][info.name].update(added=True,
                                                 server=info.server,
                                                 device_class=info._class)
            if info.name in device_mapping:
                old_server, old_instance, old_class = device_mapping[info.name]
                changes["devices"][info.name]["old_server"] = "{}/{}".format(
                    old_server, old_instance)
                changes["devices"][info.name]["old_class"] = old_class

        elif method == "delete_device":
            device, = args
            server, instance, clss = device_mapping[device]
            device_data = data["servers"][server][instance][clss]
            properties = device_data.get("properties", {})
            changes["devices"][device].update(
                deleted=True,
                server="{}/{}".format(server, instance),
                instance=instance,
                device_class=clss,
                properties=properties)

        elif method == "put_device_property":
            device, properties = args
            old_data = get_device_data(device, device_mapping, data)
            caseless_props = CaselessDict(old_data.get("properties", {}))
            if "properties" not in changes["devices"][device]:
                changes["devices"][device]["properties"] = {}
            for name, value in list(properties.items()):
                old_value = caseless_props.get(name)
                if value != old_value:
                    changes["devices"][device]["properties"].update({
                        name: {
                            "value": value,
                            "old_value": old_value
                        }
                    })

        elif method == "delete_device_property":
            device, properties = args
            old_data = get_device_data(device, device_mapping, data)
            caseless_props = CaselessDict(old_data.get("properties", {}))
            prop_changes = changes["devices"][device].setdefault(
                "properties", {})
            for prop in properties:
                old_value = caseless_props[prop]
                prop_changes[prop] = {"old_value": old_value}

        elif method == "put_device_attribute_property":
            device, properties = args
            attr_props = changes["devices"][device].setdefault(
                "attribute_properties", {})
            old_data = get_device_data(device, device_mapping, data)
            caseless_attrs = CaselessDict(old_data.get(
                "attribute_properties", {}))
            for attr, props in list(properties.items()):
                caseless_props = CaselessDict(caseless_attrs.get(attr, {}))
                for name, value in list(props.items()):
                    old_value = caseless_props.get(name)
                    if value != old_value:
                        attr_props[attr] = {
                            name: {"old_value": old_value,
                                   "value": value}
                        }

        elif method == "delete_device_attribute_property":
            device, attributes = args
            prop_changes = attr_props = changes["devices"][device].setdefault(
                "attribute_properties", {})
            old_data = get_device_data(device, device_mapping, data)
            caseless_attrs = CaselessDict(old_data.get("attribute_properties", {}))
            for attr, props in list(attributes.items()):
                caseless_props = CaselessDict(caseless_attrs[attr])
                for prop in props:
                    old_value = caseless_props.get(prop)
                    attr_props[attr] = {prop: {"old_value": old_value}}

        elif method == "put_class_property":
            clss, properties = args
            old_data = classes.get(clss, {})
            caseless_props = CaselessDict(old_data.get("properties", {}))
            prop_changes = changes["classes"][clss].setdefault("properties", {})
            for name, value in list(properties.items()):
                old_value = caseless_props.get(name)
                if value != old_value:
                    prop_changes.update({
                        name: {
                            "value": value,
                            "old_value": old_value}
                    })

        elif method == "delete_class_property":
            clss, properties = args
            old_data = classes.get(clss, {})
            caseless_props = CaselessDict(old_data.get("properties", {}))
            prop_changes = changes["classes"][clss].setdefault("properties", {})
            for prop in properties:
                old_value = caseless_props.get(prop)
                prop_changes[prop] = {"old_value": old_value}

        elif method == "put_class_attribute_property":
            clss, properties = args
            attr_props = changes["classes"][clss].setdefault(
                "attribute_properties", {})
            old_data = classes.get(clss, {})
            caseless_attrs = CaselessDict(
                old_data.get("attribute_properties", {}))
            for attr, props in list(properties.items()):
                caseless_props = CaselessDict(caseless_attrs.get(attr, {}))
                for name, value in list(props.items()):
                    old_value = caseless_props.get(name)
                    if value != old_value:
                        attr_props[attr] = {
                            name: {"old_value": old_value,
                                   "value": value}
                        }

        elif method == "delete_class_attribute_property":
            clss, attributes = args
            attr_props = changes["classes"][clss].setdefault(
                "attribute_properties", {})
            old_data = classes.get(clss, {})
            caseless_attrs = CaselessDict(old_data.get("properties", {}))
            for attr, props in list(attributes.items()):
                caseless_props = CaselessDict(caseless_attrs.get(attr, {}))
                for prop in props:
                    old_value = caseless_props.get(prop)
                    attr_props[attr] = {prop: {"old_value": old_value}}

    return {
        "devices": changes["devices"].to_dict(),
        "classes": changes["classes"].to_dict(),
    }


def show_actions(data, calls):
    """
    Print out a human readable representation of changes
    """

    changes = get_changes(data, calls)

    # Go through all the devices that have been touched
    # and print out a representation of the changes.
    # TODO: is there a more reasonable way to sort this?
    indent = " " * 2
    for device in sorted(changes["devices"]):
        info = changes["devices"][device]
        if info.get("added"):
            if info.get("old_server"):
                print(("{} Device: {}".format(yellow("="), device)))
            else:
                print(("{} Device: {}".format(green("+"), device)))
        elif info.get("deleted"):
            print(("{} Device: {}".format(red("-"), device)))
        else:
            print(("{} Device: {}".format(yellow("="), device)))

        if info.get("server"):
            if info.get("old_server"):
                print(("{}Server: {} -> {}".format(indent,
                                                   red(info["old_server"]),
                                                   green(info["server"]))))
            else:
                print(("{}Server: {}".format(indent, info["server"])))

        if info.get("device_class"):
            if info.get("old_class"):
                if info["old_class"] != info["device_class"]:
                    print(("{}Class: {} -> {}".format(indent,
                                                      info["old_class"],
                                                      info["device_class"])))
            else:
                print(("{}Class: {}".format(indent, info["device_class"])))

        if info.get("alias"):
            alias = info.get("alias").get("value")
            old_alias = info.get("alias").get("old_value")
            if old_alias:
                if old_alias != alias:
                    print(("{}Alias: {} -> {}".format(indent, red(old_alias),
                                                      green(alias))))
            else:
                print(("{}Alias: {}".format(indent, alias)))

        if info.get("properties"):
            lines = []
            for prop, change in sorted(info.get("properties", {}).items()):
                if change.get("value"):
                    value = change.get("value")
                    old_value = change.get("old_value")
                    if old_value is not None:
                        if old_value != value:
                            # change property
                            lines.append(yellow("{}= {}".format(indent * 2, prop)))
                            lines.append(property_diff(
                                change["old_value"], change["value"], indent * 3))
                    else:
                        # new property
                        lines.append(green("{}+ {}".format(indent * 2, prop)))
                        lines.append(green(format_property(change["value"],
                                                           indent * 3)))
                else:
                    # delete property
                    lines.append(red("{}- {}".format(indent * 2, prop)))
                    lines.append(red(format_property(change["old_value"], indent * 3)))
            if lines:
                print(("{}Properties:".format(indent * 1)))
                print(("\n".join(lines)))

        if info.get("attribute_properties"):
            lines = []
            for attr, props in sorted(
                    info.get("attribute_properties", {}).items()):
                attr_lines = []
                for prop, change in sorted(props.items()):
                    value = change.get("value")
                    old_value = change.get("old_value")
                    if value is not None:
                        if old_value is not None:
                            # change
                            if value != old_value:
                                attr_lines.append(yellow("{}= {}".format(indent * 3, prop)))
                                attr_lines.append(property_diff(
                                    change["old_value"], change["value"], indent * 4))
                        else:
                            # addition
                            attr_lines.append(green("{}+ {}".format(indent * 3, prop)))
                            attr_lines.append(green(format_property(change["value"],
                                                                    indent * 4)))
                    else:
                        # removal
                        attr_lines.append(red("{}- {}".format(indent * 3, prop)))
                        attr_lines.append(red(format_property(change["old_value"],
                                                              indent * 4)))
                if attr_lines:
                    lines.append("{}{}".format(indent * 2, attr))
                    lines.extend(attr_lines)
            if lines:
                print(("{}Attribute properties:".format(indent)))
                print(("\n".join(lines)))

    for clss in sorted(changes["classes"]):
        info = changes["classes"][clss]

        if info.get("properties"):
            lines = []
            for prop, change in sorted(info.get("properties", {}).items()):
                if change.get("value"):
                    value = change.get("value")
                    old_value = change.get("old_value")
                    if old_value is not None:
                        if old_value != value:
                            # change property
                            lines.append(yellow("{}= {}".format(indent * 2, prop)))
                            lines.append(property_diff(
                                change["old_value"], change["value"], indent * 3))
                    else:
                        # new property
                        lines.append(green("{}+ {}".format(indent * 2, prop)))
                        lines.append(green(format_property(change["value"],
                                                           indent * 3)))
                else:
                    # delete property
                    lines.append(red("{}- {}".format(indent * 2, prop)))
                    lines.append(red(format_property(change["old_value"], indent * 3)))
            if lines:
                print(("{} Class Properties:".format(indent * 1)))
                print(("\n".join(lines)))
        print()



"""
Functionality for configuring a Tango DB from a dsconfig file
"""
from __future__ import absolute_import
from collections import defaultdict
from functools import partial

import tango

from .appending_dict.caseless import CaselessDictionary
from .tangodb import SPECIAL_ATTRIBUTE_PROPERTIES, is_protected
from .utils import ObjectWrapper


def check_attribute_property(propname):
    # Is this too strict? Do we ever need non-standard attr props?
    if (not propname.startswith("_")
            and propname not in SPECIAL_ATTRIBUTE_PROPERTIES):
        raise KeyError("Bad attribute property name: %s" % propname)
    return True


def update_properties(db, parent, db_props, new_props,
                      attribute=False, cls=False,
                      delete=True, ignore_case=False, strict_attr_props=True):
    """
    Updates properties in DB. Covers both device and class
    properties/attribute properties.

    'parent' is the name of the containing device or class.
    """
    if ignore_case:
        caseless_db_props = CaselessDictionary(db_props)
        caseless_new_props = CaselessDictionary(new_props)
    else:
        caseless_db_props = db_props
        caseless_new_props = new_props
    # Figure out what's going to be added/changed or removed
    if attribute:
        added_props = defaultdict(dict)
        # For attribute properties we need to go one step deeper into
        # the dict, since each attribute can have several properties.
        # A little messy, but at least it's consistent.
        for attr, props in list(new_props.items()):
            for prop, value in list(props.items()):
                if ignore_case:
                    orig = CaselessDictionary(caseless_db_props.get(attr, {})).get(prop)
                else:
                    orig = caseless_db_props.get(attr, {}).get(prop)
                if value and value != orig and (not strict_attr_props
                                                or check_attribute_property(prop)):
                    added_props[attr][prop] = value
        removed_props = defaultdict(dict)
        for attr, props in list(db_props.items()):
            for prop in props:
                if ignore_case:
                    new = CaselessDictionary(new_props.get(attr, {})).get(prop)
                else:
                    new = new_props.get(attr, {}).get(prop)
                if (new is None and not is_protected(prop, True)) or new == []:
                    # empty list forces removal of "protected" properties
                    removed_props[attr][prop] = value
    else:
        added_props = {}
        for prop, value in list(new_props.items()):
            old_value = caseless_db_props.get(prop, [])
            if value and value != old_value:
                added_props[prop] = value
        removed_props = {}
        for prop, value in list(db_props.items()):
            new_value = caseless_new_props.get(prop)
            if (new_value is None and not is_protected(prop)) or new_value == []:
                # empty list forces removal of "protected" properties
                removed_props[prop] = value

    # Find the appropriate DB method to call. Thankfully the API is
    # consistent here.
    db_method_ending = (("class" if cls else "device") +
                        ("_attribute" if attribute else "") + "_property")
    put_method = getattr(db, "put_" + db_method_ending)
    delete_method = getattr(db, "delete_" + db_method_ending)

    # Do it!
    if delete and removed_props:
        delete_method(parent, removed_props)
    if added_props:
        put_method(parent, added_props)

    return added_props, removed_props


def update_server(db, server_name, server_dict, db_dict,
                  update=False, ignore_case=False,
                  difactory=tango.DbDevInfo, strict_attr_props=True):
    """
    Creates/removes devices for a given server. Optionally
    ignores removed devices, only adding new and updating old ones.
    """

    for class_name, cls in list(server_dict.items()):  # classes
        if ignore_case:
            cls = CaselessDictionary(cls)
        removed_devices = [dev for dev in db_dict.get(class_name, {})
                           if dev not in cls
                           # never remove dservers
                           and not class_name.lower() == "dserver"]
        added_devices = list(cls.items())
        if not update:
            for device_name in removed_devices:
                db.delete_device(device_name)

        for device_name, dev in added_devices:
            if ignore_case:
                devs = CaselessDictionary(db_dict.get(class_name, {}))
            else:
                devs = db_dict.get(class_name, {})
            if device_name not in devs:
                devinfo = difactory()
                devinfo.server = server_name
                devinfo._class = class_name
                devinfo.name = device_name
                db.add_device(devinfo)

            update_device(db, device_name, devs.get(device_name, {}),
                          dev, update=update, ignore_case=ignore_case,
                          strict_attr_props=strict_attr_props)

    return added_devices, removed_devices


def update_device_or_class(db, name, db_dict, new_dict,
                           cls=False, update=False, ignore_case=False,
                           strict_attr_props=True):
    """
    Configure a device or a class
    """

    # Note: if the "properties" key is missing, we'll just ignore any
    # existing properties in the DB. Ditto for attribute_properties.
    # To remove all existing properties, it should be an empty dict.

    if "properties" in new_dict:
        db_props = db_dict.get("properties", {})
        new_props = new_dict["properties"]
        update_properties(db, name, db_props, new_props, cls=cls,
                          delete=not update, ignore_case=ignore_case)

    if "attribute_properties" in new_dict:
        db_attr_props = db_dict.get("attribute_properties", {})
        new_attr_props = new_dict["attribute_properties"]
        update_properties(db, name, db_attr_props, new_attr_props,
                          attribute=True, cls=cls, delete=not update,
                          ignore_case=ignore_case,
                          strict_attr_props=strict_attr_props)

    # device aliases
    if not cls:
        if "alias" in new_dict and new_dict["alias"] != db_dict.get("alias"):
            db.put_device_alias(name, new_dict["alias"])
        elif not update and ("alias" not in new_dict and "alias" in db_dict):
            db.delete_device_alias(db_dict["alias"])


# nicer aliases
update_device = partial(update_device_or_class, cls=False)
update_class = partial(update_device_or_class, cls=True)


def configure(data, dbdata, update=False, ignore_case=False,
              strict_attr_props=True):
    """
    Takes an input data dict and the relevant current DB data.  Returns
    the DB calls needed to bring the Tango DB to the state described
    by 'data'.  The 'update' flag means that servers/devices are not
    removed, only added or changed. If the 'ignore_case' flag is True,
    the names of servers, devices and properties will be treated as
    caseless.

    Note: This function does *not* itself modify the Tango DB. It passes a
    "fake" database object around that just records what the various other
    functions do to it, and then returns the list of calls made.
    """

    db = ObjectWrapper()
    if ignore_case:
        dbdata = CaselessDictionary(dbdata)
    for servername, serverdata in list(data.get("servers", {}).items()):
        for instname, instdata in list(serverdata.items()):
            dbinstdata = (dbdata.get("servers", {})
                          .get(servername, {})
                          .get(instname, {}))
            added, removed = update_server(
                db, "%s/%s" % (servername, instname),
                instdata, dbinstdata, update, ignore_case,
                strict_attr_props=strict_attr_props)

    for classname, classdata in list(data.get("classes", {}).items()):
        dbclassdata = dbdata.get("classes", {}).get(classname, {})
        update_class(db, classname, dbclassdata, classdata, update=update)

    return db.calls



from __future__ import absolute_import
from __future__ import print_function
import json
import sys
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

from .utils import green, yellow, red


def decode_pointer(ptr):
    """
    Take a string representing a JSON pointer and return a
    sequence of parts, decoded.
    """
    return [p.replace("~1", "/").replace("~0", "~")
            for p in ptr.split("/")]


def dump_value(value):
    """
    Make a string out of a value, for printing
    """
    if value is not None:
        if isinstance(value, Mapping):
            dump = json.dumps(value, indent=4)
            return dump
        return repr(value)
    else:
        return "None"  # should never happen?


def print_diff(dbdict, data, removes=True):
    """
    Print a (hopefully) human readable list of changes.
    """

    # TODO: needs work, especially on multiline properties,
    # empty properties (should probably never be allowed but still)
    # and probably more corner cases. Also the output format could
    # use some tweaking.

    try:
        from collections import defaultdict
        import jsonpatch
        from jsonpointer import resolve_pointer, JsonPointerException

        ops = defaultdict(int)

        diff = jsonpatch.make_patch(dbdict, data)
        for d in diff:
            try:
                ptr = " > ".join(decode_pointer(d["path"]))
                if d["op"] == "replace":
                    print(yellow("REPLACE:"))
                    print(yellow(ptr))
                    db_value = resolve_pointer(dbdict, d["path"])
                    print(red(dump_value(db_value)))
                    print(green(dump_value(d["value"])))
                    ops["replace"] += 1
                if d["op"] == "add":
                    print(green("ADD:"))
                    print(green(ptr))
                    if d["value"]:
                        print(green(dump_value(d["value"])))
                    ops["add"] += 1
                if removes and d["op"] == "remove":
                    print(red("REMOVE:"))
                    print(red(ptr))
                    value = resolve_pointer(dbdict, d["path"])
                    if value:
                        print(red(dump_value(value)))
                    ops["remove"] += 1
            except JsonPointerException as e:
                print(" - Error parsing diff - report this!: %s" % e)
        # # The following output is a bit misleading, removing for now
        # print "Total: %d operations (%d replace, %d add, %d remove)" % (
        #     sum(ops.values()), ops["replace"], ops["add"], ops["remove"])
        return diff
    except ImportError:
        print(("'jsonpatch' module not available - "
               "no diff printouts for you! (Try -d instead.)"), file=sys.stderr)



"""
A simple terminal viewer for JSON files. It supports folding
and traversal. Currently does not support arbitrary JSON
but should handle dsconfig files.

Useful additions might include search, filtering and possibly
editing features.
"""

import urwid
from urwidtrees.decoration import CollapsibleIndentedTree
from urwidtrees.tree import Tree
from urwidtrees.widgets import TreeBox


class FocusableText(urwid.WidgetWrap):

    def __init__(self, path, data):
        txt = path[-1]
        child = get_path(path, data)
        if isinstance(txt, int):  # we're in a list
            # TODO: support containers inside lists
            a = urwid.Text(str(txt) + ":")
            if isinstance(child, dict):
                b = urwid.Text("<object>")
            else:
                b = urwid.Text(child)
            t = urwid.Columns([("pack", a), b], dividechars=2)
        else:
            if isinstance(child, (dict, list)):
                t = urwid.Text(str(txt))
            else:
                t = urwid.Text("%s:  %s" % (txt, child))
        w = urwid.AttrMap(t, 'body', 'focus')
        urwid.WidgetWrap.__init__(self, w)

    def selectable(self):
        return True

    def keypress(self, size, key):
        return key


def get_path(path, dictlike):
    for step in path:
        dictlike = dictlike[step]
    return dictlike


def del_path(path, dictlike):
    for step in path[:-1]:
        dictlike = dictlike[step]
    del dictlike[path[-1]]


class MyTree(Tree):

    def __init__(self, data):
        self.data = data
        self.root = tuple(data.keys())

    def __getitem__(self, pos):
        return FocusableText(pos, self.data)

    # helpers

    def _get_children(self, path):
        node = get_path(path, self.data)
        if node:
            if isinstance(node, dict):
                children = list(node.keys())
                return [path + (child,) for child in sorted(children)]
            elif isinstance(node, list):
                children = list(range(len(node)))
                return [path + (child,) for child in children]
        return []

    def _get_siblings(self, path):
        parent = self.parent_position(path)
        if parent is not None:
            return self._get_children(parent)
        return [path]

    # UrwidTrees Tree API

    def parent_position(self, path):
        if path != self.root:
            return path[:-1]

    def first_child_position(self, path):
        try:
            children = self._get_children(path)
        except KeyError:
            return path
        if children:
            return children[0]

    def last_child_position(self, path):
        try:
            children = self._get_children(path)
        except KeyError:
            return path
        if children:
            return children[-1]

    def next_sibling_position(self, path):
        siblings = self._get_siblings(path)
        if path in siblings:
            myindex = siblings.index(path)
            if myindex + 1 < len(siblings):  # path is not the last entry
                return siblings[myindex + 1]

    def prev_sibling_position(self, path):
        siblings = self._get_siblings(path)
        if path in siblings:
            myindex = siblings.index(path)
            if myindex > 0:  # path is not the first entry
                return siblings[myindex - 1]


class MyTreeBox(TreeBox):

    def keypress(self, size, key):
        """
        Spicing up the keybindings!
        """
        # if key == "delete":
        #     _, path = self.get_focus()
        #     self.set_focus(path[:-1])
        #     del_path(path, treedata)
        #     self.refresh()
        if key == "enter":
            w, pos = self._walker.get_focus()
            if self._tree.is_collapsed(pos):
                self.expand_focussed()
            else:
                self.collapse_focussed()
        elif key == "right":
            self.expand_focussed()
            self.focus_first_child()
        elif key in ("q", "esc"):
            raise urwid.ExitMainLoop()
        else:
            return TreeBox.keypress(self, size, key)


palette = [
    ('focus', 'light gray', 'dark blue', 'standout'),
    ('bars', 'dark blue', 'light gray', ''),
    ('arrowtip', 'light blue', 'light gray', ''),
    ('connectors', 'light red', 'light gray', ''),
]

if __name__ == "__main__":

    import json
    import signal
    import sys

    # make ctrl+C exit properly, without breaking the terminal
    def exit_handler(signum, frame):
        raise urwid.ExitMainLoop()


    signal.signal(signal.SIGINT, exit_handler)

    # redirecting stdin breaks things. This was supposed to help
    # but it does not. Fixme!
    old_stdin = sys.stdin
    sys.stdin = open('/dev/tty')
    screen = urwid.raw_display.Screen()

    sys.stdin = old_stdin
    if len(sys.argv) == 2:
        with open(sys.argv[1]) as f:
            treedata = {sys.argv[1]: json.load(f)}
    else:
        # # try to read from stdin
        # treedata = {"stdin": json.load(sys.stdin)}
        sys.exit("You must give a JSON file as input")

    tree = CollapsibleIndentedTree(MyTree(treedata), indent=2,
                                   is_collapsed=lambda path: len(path) > 1,
                                   arrow_tip_char=None,
                                   icon_frame_left_char=None,
                                   icon_frame_right_char=None,
                                   icon_collapsed_char="+",
                                   icon_expanded_char="-")
    box = MyTreeBox(tree)
    loop = urwid.MainLoop(box, palette, screen=screen)
    loop.run()



"""
Reads a JSON file in the right format, compares it with the current
state of the Tango DB, and generates the set of DB API commands needed
to get to the state described by the file. These commands can also
optionally be run.
"""
from __future__ import absolute_import
from __future__ import print_function

import json
import sys
import time
from optparse import OptionParser
from tempfile import NamedTemporaryFile

import tango
from dsconfig.appending_dict.caseless import CaselessDictionary
from dsconfig.configure import configure
from dsconfig.dump import get_db_data
from dsconfig.filtering import filter_config
from dsconfig.formatting import (CLASSES_LEVELS, SERVERS_LEVELS, load_json,
                                 normalize_config, validate_json,
                                 clean_metadata)
from dsconfig.output import show_actions
from dsconfig.tangodb import summarise_calls, get_devices_from_dict
from dsconfig.utils import SUCCESS, ERROR, CONFIG_APPLIED, CONFIG_NOT_APPLIED
from dsconfig.utils import green, red, yellow, progressbar, no_colors


def json_to_tango(options, args):

    if options.no_colors:
        no_colors()

    if len(args) == 0:
        data = load_json(sys.stdin)
    else:
        json_file = args[0]
        with open(json_file) as f:
            data = load_json(f)

    # Normalization - making the config conform to standard
    data = normalize_config(data)

    # remove any metadata at the top level (should we use this for something?)
    data = clean_metadata(data)

    # Optional validation of the JSON file format.
    if options.validate:
        validate_json(data)

    # filtering
    try:
        if options.include:
            data["servers"] = filter_config(
                data.get("servers", {}), options.include, SERVERS_LEVELS)
        if options.exclude:
            data["servers"] = filter_config(
                data.get("servers", {}), options.exclude, SERVERS_LEVELS, invert=True)
        if options.include_classes:
            data["classes"] = filter_config(
                data.get("classes", {}), options.include_classes, CLASSES_LEVELS)
        if options.exclude_classes:
            data["classes"] = filter_config(
                data.get("classes", {}), options.exclude_classes, CLASSES_LEVELS,
                invert=True)
    except ValueError as e:
        print(red("Filter error:\n%s" % e), file=sys.stderr)
        sys.exit(ERROR)

    if not any(k in data for k in ("devices", "servers", "classes")):
        sys.exit(ERROR)

    if options.input:
        print(json.dumps(data, indent=4))
        return

    # check if there is anything in the DB that will be changed or removed
    db = tango.Database()
    if options.dbdata:
        with open(options.dbdata) as f:
            original = json.loads(f.read())
        collisions = {}
    else:
        original = get_db_data(db, dservers=True, class_properties=True)
        if "servers" in data:
            devices = CaselessDictionary({
                dev: (srv, inst, cls)
                for srv, inst, cls, dev
                in get_devices_from_dict(data["servers"])
            })
        else:
            devices = CaselessDictionary({})
        orig_devices = CaselessDictionary({
            dev: (srv, inst, cls)
            for srv, inst, cls, dev
            in get_devices_from_dict(original["servers"])
        })
        collisions = {}
        for dev, (srv, inst, cls) in list(devices.items()):
            if dev in orig_devices:
                server = "{}/{}".format(srv, inst)
                osrv, oinst, ocls = orig_devices[dev]
                origserver = "{}/{}".format(osrv, oinst)
                if server.lower() != origserver.lower():
                    collisions.setdefault(origserver, []).append((ocls, dev))

    # get the list of DB calls needed
    dbcalls = configure(data, original,
                        update=options.update,
                        ignore_case=not options.case_sensitive,
                        strict_attr_props=not options.nostrictcheck)

    # Print out a nice diff
    if options.verbose:
        show_actions(original, dbcalls)

    # perform the db operations (if we're supposed to)
    if options.write and dbcalls:
        for i, (method, args, kwargs) in enumerate(dbcalls):
            if options.sleep:
                time.sleep(options.sleep)
            if options.verbose:
                progressbar(i, len(dbcalls), 20)
            getattr(db, method)(*args, **kwargs)
        print()

    # optionally dump some information to stdout
    if options.output:
        print(json.dumps(original, indent=4))
    if options.dbcalls:
        print("Tango database calls:", file=sys.stderr)
        for method, args, kwargs in dbcalls:
            print(method, args, file=sys.stderr)

    # Check for moved devices and remove empty servers
    empty = set()
    for srvname, devs in list(collisions.items()):
        if options.verbose:
            srv, inst = srvname.split("/")
            for cls, dev in devs:
                print(red("MOVED (because of collision):"), dev, file=sys.stderr)
                print("    Server: ", "{}/{}".format(srv, inst), file=sys.stderr)
                print("    Class: ", cls, file=sys.stderr)
        if len(db.get_device_class_list(srvname)) == 2:  # just dserver
            empty.add(srvname)
            if options.write:
                db.delete_server(srvname)

    # finally print out a brief summary of what was done
    if dbcalls:
        print()
        print("Summary:", file=sys.stderr)
        print("\n".join(summarise_calls(dbcalls, original)), file=sys.stderr)
        if collisions:
            servers = len(collisions)
            devices = sum(len(devs) for devs in list(collisions.values()))
            print(red("Move %d devices from %d servers." %
                      (devices, servers)), file=sys.stderr)
        if empty and options.verbose:
            print(red("Removed %d empty servers." % len(empty)), file=sys.stderr)

        if options.write:
            print(red("\n*** Data was written to the Tango DB ***"), file=sys.stderr)
            with NamedTemporaryFile(prefix="dsconfig-", suffix=".json",
                                    delete=False) as f:
                f.write(json.dumps(original, indent=4).encode())
                print(("The previous DB data was saved to %s" %
                       f.name), file=sys.stderr)
            sys.exit(CONFIG_APPLIED)
        else:
            print(yellow(
                "\n*** Nothing was written to the Tango DB (use -w) ***"), file=sys.stderr)
            sys.exit(CONFIG_NOT_APPLIED)

    else:
        print(green("\n*** No changes needed in Tango DB ***"), file=sys.stderr)
        sys.exit(SUCCESS)


def main():

    usage = "Usage: %prog [options] JSONFILE"
    parser = OptionParser(usage=usage)

    parser.add_option("-w", "--write", dest="write", action="store_true",
                      help="write to the Tango DB", metavar="WRITE")
    parser.add_option("-u", "--update", dest="update", action="store_true",
                      help="don't remove things, only add/update",
                      metavar="UPDATE")
    parser.add_option("-c", "--case-sensitive", dest="case_sensitive",
                      action="store_true",
                      help=("Don't ignore the case of server, device, "
                            "attribute and property names"),
                      metavar="CASESENSITIVE")
    parser.add_option("-q", "--quiet",
                      action="store_false", dest="verbose", default=True,
                      help="don't print actions to stderr")
    parser.add_option("-o", "--output", dest="output", action="store_true",
                      help="Output the relevant DB state as JSON.")
    parser.add_option("-p", "--input", dest="input", action="store_true",
                      help="Output the input JSON (after filtering).")
    parser.add_option("-d", "--dbcalls", dest="dbcalls", action="store_true",
                      help="print out all db calls.")
    parser.add_option("-v", "--no-validation", dest="validate", default=True,
                      action="store_false", help=("Skip JSON validation"))
    parser.add_option("-s", "--sleep", dest="sleep", default=0.01,
                      type="float",
                      help=("Number of seconds to sleep between DB calls"))
    parser.add_option("-n", "--no-colors",
                      action="store_true", dest="no_colors", default=False,
                      help="Don't print colored output")
    parser.add_option("-i", "--include", dest="include", action="append",
                      help=("Inclusive filter on server configutation"))
    parser.add_option("-x", "--exclude", dest="exclude", action="append",
                      help=("Exclusive filter on server configutation"))
    parser.add_option("-a", "--no-strict-check", dest="nostrictcheck",
                      default=False, action="store_true",
                      help="Disable strick attribute property checking")
    parser.add_option("-I", "--include-classes", dest="include_classes",
                      action="append",
                      help=("Inclusive filter on class configuration"))
    parser.add_option("-X", "--exclude-classes", dest="exclude_classes",
                      action="append",
                      help=("Exclusive filter on class configuration"))

    parser.add_option(
        "-D", "--dbdata",
        help="Read the given file as DB data instead of using the actual DB",
        dest="dbdata")

    options, args = parser.parse_args()

    json_to_tango(options, args)


if __name__ == "__main__":
    main()



import sys
from functools import partial

# exit codes
SUCCESS = 0  # NO DB CHANGES
ERROR = 1
CONFIG_APPLIED = 2
CONFIG_NOT_APPLIED = 3

# colors
ADD = GREEN = '\033[92m'
REMOVE = RED = FAIL = '\033[91m'
REPLACE = YELLOW = WARN = '\033[93m'
ENDC = '\033[0m'


def no_colors():
    global ADD, GREEN, REMOVE, RED, FILE, REPLACE, YELLOW, WARN, ENDC
    ADD = ''
    GREEN = ''
    REMOVE = ''
    RED = ''
    FAIL = ''
    REPLACE = ''
    YELLOW = ''
    WARN = ''
    ENDC = ''


def green(text):
    return GREEN + text + ENDC


def red(text):
    return RED + text + ENDC


def yellow(text):
    return YELLOW + text + ENDC


def progressbar(i, n, width):
    if n == 1:
        progress = 1.0  # done!
    else:
        progress = float(i) / (n - 1)
    hashes = '#' * int(round(progress * width))
    spaces = ' ' * (width - len(hashes))
    sys.stdout.write(
        "\rProgress: [{0}] {1}%".format(hashes + spaces,
                                        int(round(progress * 100))))
    sys.stdout.flush()


def find_device(definitions, devname, caseless=False):
    """
    Find a given device in a server dict
    """
    for srvname, srv in list(definitions["servers"].items()):
        if caseless:
            srv = CaselessDict(srv)
        for instname, inst in list(srv.items()):
            if caseless:
                inst = CaselessDict(inst)
            for classname, cls in list(inst.items()):
                if caseless:
                    cls = CaselessDict(cls)
                if devname in cls:
                    return cls[devname], (srvname, instname, classname, devname)
    raise ValueError("device '%s' not defined" % devname)


def find_class(definitions, clsname):
    """
    Find a given device in a server dict
    """
    for instname, inst in list(definitions["servers"].items()):
        if clsname in inst:
            return inst[clsname]
    raise ValueError("class '%s' not defined" % clsname)


def get_devices_from_dict(dbdict):
    return [(server_name, inst_name, class_name, device_name)
            for server_name, server in list(dbdict.items())
            for inst_name, inst in list(server.items())
            for class_name, clss in list(inst.items())
            for device_name in clss]


class ObjectWrapper(object):
    """
    An object that allows all method calls and records them,
    then passes them on to a target object (if any).
    """

    def __init__(self, target=None):
        self.target = target
        self.calls = []

    def __getattr__(self, attr):
        def method(attr, *args, **kwargs):
            self.calls.append((attr, args, kwargs))
            if self.target:
                getattr(self.target, attr)(*args, **kwargs)

        return partial(method, attr)


# From http://code.activestate.com/recipes/66315/#c8

#  27-05-04
# v2.0.2
#

# caseless
# Featuring :

# caselessDict
# A case insensitive dictionary that only permits strings as keys.

# Implemented for ConfigObj
# Requires Python 2.2 or above

# Copyright Michael Foord
# Not for use in commercial projects without permission. (Although permission will probably be given).
# If you use in a non-commercial project then please credit me and include a link back.
# If you release the project non-commercially then let me know (and include this message with my code !)

# No warranty express or implied for the accuracy, fitness to purpose or otherwise for this code....
# Use at your own risk !!!

# E-mail fuzzyman AT atlantibots DOT org DOT uk (or michael AT foord DOT me DOT uk )
# Maintained at www.voidspace.org.uk/atlantibots/pythonutils.html


class CaselessDict(dict):
    """
    A case insensitive dictionary that only permits strings as keys.
    """

    def __init__(self, indict={}):
        dict.__init__(self)
        # not self.__keydict because I want it to be easily accessible by subclasses
        self._keydict = {}
        for entry in indict:
            # not dict.__setitem__(self, entry, indict[entry]) becasue this causes errors
            # (phantom entries) where indict has overlapping keys...
            self[entry] = indict[entry]

    def findkey(self, item):
        """
        A caseless way of checking if a key exists or not.
        It returns None or the correct key.
        """
        if not isinstance(item, str): raise TypeError(
            'Keywords for this object must be strings. You supplied %s' % type(item))
        key = item.lower()
        try:
            return self._keydict[key]
        except:
            return None

    def changekey(self, item):
        """
        For changing the casing of a key.
        If a key exists that is a caseless match for 'item' it will be changed to 'item'.
        This is useful when initially setting up default keys - but later might want to
        preserve an alternative casing. (e.g. if later read from a config file - and you
         might want to write back out with the user's casing preserved).
        """
        key = self.findkey(item)  # does the key exist
        if key == None: raise KeyError(item)
        temp = self[key]
        del self[key]
        self[item] = temp
        self._keydict[item.lower()] = item

    def lowerkeys(self):
        """
        Returns a lowercase list of all member keywords.
        """
        return list(self._keydict.keys())

    def __setitem__(self, item, value):  # setting a keyword
        """
        To implement lowercase keys.
        """
        key = self.findkey(item)  # if the key already exists
        if key != None:
            dict.__delitem__(self, key)
        self._keydict[item.lower()] = item
        dict.__setitem__(self, item, value)

    def __getitem__(self, item):
        """
        To implement lowercase keys.
        """
        key = self.findkey(item)  # does the key exist
        if key == None: raise KeyError(item)
        return dict.__getitem__(self, key)

    def __delitem__(self, item):  # deleting a keyword
        key = self.findkey(item)  # does the key exist
        if key == None: raise KeyError(item)
        dict.__delitem__(self, key)
        del self._keydict[item.lower()]

    def pop(self, item, default=None):
        """
        Correctly emulates the pop method.
        """
        key = self.findkey(item)  # does the key exist
        if key == None:
            if default == None:
                raise KeyError(item)
            else:
                return default
        del self._keydict[item.lower()]
        return dict.pop(self, key)

    def popitem(self):
        """
        Correctly emulates the popitem method.
        """
        popped = dict.popitem(self)
        del self._keydict[popped[0].lower()]
        return popped

    def has_key(self, item):
        """
        A case insensitive test for keys.
        """
        if not isinstance(item, str): return False  # should never have a non-string key
        return item.lower() in self._keydict  # does the key exist

    def __contains__(self, item):
        """
        A case insensitive __contains__.
        """
        if not isinstance(item, str): return False  # should never have a non-string key
        return item.lower() in self._keydict  # does the key exist

    def setdefault(self, item, default=None):
        """
        A case insensitive setdefault.
        If no default is supplied it sets the item to None
        """
        key = self.findkey(item)  # does the key exist
        if key != None: return self[key]
        self.__setitem__(item, default)
        self._keydict[item.lower()] = item
        return default

    def get(self, item, default=None):
        """A case insensitive get."""
        key = self.findkey(item)  # does the key exist
        if key != None: return self[key]
        return default

    def update(self, indict):
        """
        A case insensitive update. If your dictionary has overlapping keys (e.g. 'FISH'
        and 'fish') then one will overwrite the other. The one that is kept is arbitrary.
        """
        for entry in indict:
            self[entry] = indict[entry]  # this uses the new __setitem__ method

    def copy(self):
        """
        Create a new caselessDict object that is a copy of this one.
        """
        return CaselessDict(self)

    def dict(self):
        """
        Create a dictionary version of this caselessDict.
        """
        return dict.copy(self)

    def clear(self):
        """
        Clear this caselessDict.
        """
        self._keydict = {}
        dict.clear(self)

    def __repr__(self):
        """A caselessDict version of __repr__ """
        return 'caselessDict(' + dict.__repr__(self) + ')'


"""
A tuctionary, or tuct, is the combination of a tuple with
a dictionary. A tuct has named items, but they cannot be
deleted or rebound, nor new can be added.
"""


class ImmutableDict(object):
    """
    The tuct class. An immutable dictionary.
    """

    def __init__(self, dict=None, **kwds):
        self.__data = {}
        if dict is not None:
            self.__data.update(dict)
        if len(kwds):
            self.__data.update(kwds)

    # del __init__

    def __repr__(self):
        return repr(self.__data)

    def __cmp__(self, dict):
        if isinstance(dict, ImmutableDict):
            return cmp(self.__data, dict.__data)
        else:
            return cmp(self.__data, dict)

    def __len__(self):
        return len(self.__data)

    def __getitem__(self, key):
        return self.__data[key]

    def copy(self):
        if self.__class__ is ImmutableDict:
            return ImmutableDict(self.__data.copy())
        __data = self.__data



import io
from setuptools import setup

with io.open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

tests_require = ["pytest", "pytest-cov", "Faker", "mock"]

setup(
    # Package
    name="dsconfig",
    version="1.6.5",
    author="KITS",
    author_email="kits-sw@maxiv.lu.se",
    url="https://gitlab.com/MaxIV/lib-maxiv-dsconfig",
    license="GPLv3",
    packages=["dsconfig", "dsconfig.appending_dict"],
    description="Library and utilities for Tango device configuration.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="Tango",
    # Requirements
    python_requires=">=2.7",
    install_requires=["jsonpatch>=1.13", "jsonschema", "six", "xlrd", "pytango"],
    extras_require={"tests": tests_require},
    # Resources
    package_data={"dsconfig": ["schema/dsconfig.json", "schema/schema2.json"]},
    # Scripts
    entry_points={
        "console_scripts": [
            "xls2json = dsconfig.excel:main",
            "json2tango = dsconfig.json2tango:main",
        ]
    },
)



from __future__ import absolute_import
from collections import defaultdict
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

from .caseless import CaselessDictionary


class SetterDict(CaselessDictionary, defaultdict):
    """
    A recursive defaultdict with extra bells & whistles

    It enables you to set keys at any depth without first creating the
    intermediate levels, e.g.

      d = SetterDict()
      d["a"]["b"]["c"] = 1

    It also allows access using normal getattr syntax, interchangeably:

      d.a["b"].c == d.a.b.c == d["a"]["b"]["c"]

    Note: only allows string keys for now.

    Keys are caseless, meaning that the key "MyKey" is the same as
    "mykey", "MYKEY", etc. The case

    Note: this thing is dangerous! Accessing a non-existing key will
    result in creating it, which means that confusing behavior is
    likely. Please use it carefully and convert to an ordinary dict
    (using to_dict()) when you're done creating it.
    """

    def __init__(self, value={}, factory=None):
        factory = factory or SetterDict
        self.__dict__["_factory"] = factory
        CaselessDictionary.__init__(self)
        defaultdict.__init__(self, factory)
        for k, v in list(value.items()):
            self[k] = v

    def __getitem__(self, key):
        try:
            return CaselessDictionary.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __setitem__(self, key, value):
        if isinstance(value, SetterDict):
            CaselessDictionary.__setitem__(self, key, value)
        elif isinstance(value, Mapping):
            CaselessDictionary.__setitem__(self, key, self._factory(value))
        else:
            CaselessDictionary.__setitem__(self, key, value)

    def __getattr__(self, name):
        return self.__getitem__(name)

    def __setattr__(self, key, value):
        return self.__setitem__(key, value)

    def to_dict(self):
        """
        Returns a ordinary dict version of itself
        """
        result = {}
        for key, value in list(self.items()):
            if isinstance(value, SetterDict):
                result[key] = value.to_dict()
            else:
                result[key] = value
        return result


def merge(d, u):
    """
    Recursively 'merge' a Mapping into another
    """
    for k, v in list(u.items()):
        if isinstance(v, Mapping):
            if k in d:
                merge(d[k], v)
            else:
                d[k] = v
        elif isinstance(d, Mapping):
            d[k] = u[k]


def list_of_strings(value):
    if isinstance(value, (list, tuple)):
        return [str(v) for v in value]
    else:
        return [str(value)]


class AppendingDict(SetterDict):
    """
    An extra weird SetterDict where assignment adds items instead of
    overwriting. It also allows setting nested values using dicts (or
    any Mapping). Scalar values are converted into lists of strings.

    Plus of course the features of SetterDict.

    !!!Attention!!!  This is pretty dangerous. It is only intended for
    cases where you need to do a lot of updating and don't want the
    hassle of creating a million recursive dicts. It will silently
    convert things and it will create any paths that are accessed
    (even without writing), etc. It is very risky to use it like a
    normal dict. Always convert it to a dict using to_dict() when
    you're done changing it! You can always create an AppendingDict
    from it again if needed later.

    a = AppendingDict()
    a.b = 1
    a.b
    -> ['1']
    a.b = 2
    a.b
    -> ['1', '2']

    a = AppendingDict()
    a.b.c.d = 3
    a.b = {"c": {"d": ['3']}}
    a
    -> {"b": {"c": {"d": ['3', '4']}}}

    """

    def __init__(self, value={}):
        SetterDict.__init__(self, value, AppendingDict)

    def _set(self, attr, value):
        SetterDict.__setitem__(self, attr, value)

    def __setitem__(self, attr, value):
        # I apologize for this method :(
        if attr in self:
            if isinstance(self[attr], AppendingDict):
                if isinstance(value, Mapping):
                    merge(self[attr], value)
                else:
                    raise ValueError("Can't overwrite a subtree "
                                     "with a scalar value.")
            else:
                self[attr].extend(list_of_strings(value))
        else:
            if isinstance(value, Mapping):
                if isinstance(value, AppendingDict):
                    self._set(attr, value)
                else:
                    self._set(attr, AppendingDict(value))
            else:
                self._set(attr, list_of_strings(value))



from __future__ import absolute_import
try:
    import unittest2 as unittest
except:
    import unittest
import pytest

from . import SetterDict, AppendingDict, merge


class MergeTestCase(unittest.TestCase):

    def test_merge(self):
        a = {1: 2, 3: 4}
        b = {5: 6}
        merge(a, b)
        self.assertEqual(a, {1: 2, 3: 4, 5: 6})

    def test_merge_mapping(self):
        a = {1: 2}
        b = {3: {4: 5}}
        merge(a, b)
        self.assertEqual(a, {1: 2, 3: {4: 5}})


class SetterDictTestCase(unittest.TestCase):

    def test_init_tiny(self):
        TINY_DICT = {"a": 1}
        sd = SetterDict(TINY_DICT)
        self.assertDictEqual(sd.to_dict(), TINY_DICT)

    def test_init_flat(self):
        FLAT_DICT = {"a": 1, "b": 2, "c": 3}
        sd = SetterDict(FLAT_DICT)
        self.assertDictEqual(sd.to_dict(), FLAT_DICT)

    def test_init_nested(self):
        NESTED_DICT = {"a": {"b": 2}}
        sd = SetterDict(NESTED_DICT)
        self.assertDictEqual(sd.to_dict(), NESTED_DICT)

    def test_init_blended(self):
        BLENDED_DICT = {"a": 1, "b": {"c": 3}}
        sd = SetterDict(BLENDED_DICT)
        self.assertDictEqual(sd.to_dict(), BLENDED_DICT)

    def test_init_deep(self):
        DEEP_DICT = {"a": {"b": {"c": {"d": 4}}}}
        sd = SetterDict(DEEP_DICT)
        self.assertDictEqual(sd.to_dict(), DEEP_DICT)

    def test_init_deep(self):
        COMPLEX_DICT = {"a": {"b": {"c": {"d": 4}}}, "e": 5}
        sd = SetterDict(COMPLEX_DICT)
        self.assertDictEqual(sd.to_dict(), COMPLEX_DICT)

    def test_setting(self):
        sd = SetterDict()
        sd["foo"] = 1
        self.assertDictEqual(sd.to_dict(), {"foo": 1})

    def test_setting_nested(self):
        sd = SetterDict()
        sd["foo"]["bar"] = 2
        self.assertDictEqual(sd.to_dict(), {"foo": {"bar": 2}})

    def test_setting_nested_nonempty(self):
        sd = SetterDict({"a": 1})
        sd["foo"]["bar"] = 2
        self.assertDictEqual(sd.to_dict(), {"a": 1, "foo": {"bar": 2}})

    def test_setting_attr(self):
        sd = SetterDict({"a": 1})
        sd.a = 2
        self.assertDictEqual(sd.to_dict(), {"a": 2})

    def test_setting_attr_deep(self):
        sd = SetterDict()
        sd.a.b.c = 4
        self.assertDictEqual(sd.to_dict(), {"a": {"b": {"c": 4}}})

    def test_to_dict(self):
        orig = {"a": {"b": ["3"], "c": {"d": ["4"]}, "e": ["1"]}}
        sd = SetterDict(orig)
        d = sd.to_dict()
        self.assertDictEqual(orig, d)

    def test_keys_case_insensitive(self):
        sd = SetterDict()
        sd.a.B.c = 1
        self.assertEqual(sd.A.b.C, sd.a.b.c, 1)

    def test_keeps_original_key_case(self):
        sd = SetterDict()
        sd.FoO = 1
        sd.foo = 2
        sd.baR = 3
        sd.BAR = 4
        self.assertListEqual(list(sd.keys()), ["FoO", "baR"])


class AppendingDictTestCase(unittest.TestCase):

    def test_basic_appending(self):
        ad = AppendingDict()
        ad["a"] = 1
        self.assertDictEqual(ad.to_dict(), {"a": ['1']})
        ad["a"] = 2
        self.assertDictEqual(ad.to_dict(), {"a": ['1', '2']})

    def test_deep_appending(self):
        ad = AppendingDict()
        ad["a"]["b"]["c"] = 1
        ad["a"]["b"]["c"] = 2
        print((type(ad["a"]["b"])))
        self.assertDictEqual(ad.to_dict(), {"a": {"b": {"c": ['1', '2']}}})

    def test_initial_setting_with_dict(self):
        ad = AppendingDict()
        ad.a = {"b": {"c": 1}}
        self.assertDictEqual(ad.to_dict(), {"a": {"b": {"c": ["1"]}}})

    def test_deep_setting_with_dict(self):
        ad = AppendingDict()
        ad.a.b.c = 1
        ad.a = {"b": {"d": 2}}
        self.assertDictEqual(ad.to_dict(), {"a": {"b": {"c": ['1'], "d": ['2']}}})

    def test_setting_with_sequence(self):
        ad = AppendingDict()
        ad.a = [1, "b"]
        self.assertDictEqual(ad.to_dict(), {"a": ['1', 'b']})

    def test_setting_existing_key_with_sequence(self):
        ad = AppendingDict()
        ad.a = [1, "b"]
        ad.a = [2, "c"]
        self.assertDictEqual(ad.to_dict(), {"a": ['1', 'b', '2', 'c']})

    def test_error_setting_existing_subtree_with_scalar(self):
        ad = AppendingDict()
        ad.a.b.c = 1

        with pytest.raises(ValueError):
            ad.a = 2

    def test_setting_with_appendingdict(self):
        ad = AppendingDict()
        ad2 = AppendingDict({"b": 3, "c": {"d": 4}})
        ad.a.e = 1
        ad.a = ad2
        self.assertDictEqual(
            ad.to_dict(), {"a": {"b": ["3"], "c": {"d": ["4"]}, "e": ["1"]}})

    def test_updating(self):
        ad = AppendingDict()
        d = {"a": 1, "b": {"c": 3}}
        ad.update(d)
        assert ad.to_dict() == {"a": ["1"], "b": {"c": ["3"]}}

    def test_set_string_value(self):
        ad = AppendingDict()
        ad.a = "abc"
        self.assertDictEqual(ad.to_dict(), {"a": ["abc"]})

    def test_to_dict(self):
        orig = {"a": {"b": ["3"], "c": {"d": ["4"]}, "e": ["1"]}}
        ad = AppendingDict(orig)
        d = ad.to_dict()
        self.assertDictEqual(orig, d)
